#!/usr/bin/env python

import sys

if sys.version_info<(3,0,0):
	import pygtk as PyGtk
	import gtk   as Gtk
	import gtk   as GLib
else:
	from gi.repository import Gtk
	from gi.repository import GLib

if sys.version_info<(3,0,0):
	ICON_SIZE_BUTTON = Gtk.ICON_SIZE_BUTTON
else:
	ICON_SIZE_BUTTON = Gtk.IconSize.BUTTON

class Visual:
	def __init__(self, config, device, logger):
		def get_objects(names):
			return list(map(lambda x:
				self.builder.get_object(x), names))

		self.config   = config
		self.device   = device
		self.logger   = logger
		self.builder  = Gtk.Builder()
		self.history  = [0.5]*1000

		self.builder.add_from_file('visual.ui')
		self.builder.connect_signals(self)
		self.window   = self.builder.get_object("window")
		self.settings = self.builder.get_object("settings")

		self.accs     = get_objects(['accx', 'accy', 'accz'])
		self.mags     = get_objects(['magx', 'magy', 'magz'])
		self.lgts     = get_objects(['lgt'])
		self.tchs     = get_objects(['tchp', 'tchd'])
		self.a2ds     = get_objects(['a2d0', 'a2d1', 'a2d2', 'a2d3', 'a2d4', 'a2d5'])

	# Signal handlers
	def on_hide(self, win, *args):
		self.settings.hide()
		return True

	def on_show(self, act):
		self.settings.show()
		return True

	def on_serial(self, act):
		if act.get_active():
			self.status(self.device.connect())
		else:
			self.status(self.device.disconnect())
		return True

	def on_xively(self, act):
		if act.get_active():
			self.status(self.logger.connect())
		else:
			self.status(self.logger.disconnect())
		return True

	def on_flush(self, act):
		self.logger.flush()

	def on_enable(self, obj):
		name  = Gtk.Buildable.get_name(obj)
		name  = name.replace('_btn', '')
		state = obj.get_active()
		if self.config.enable[name] != state:
			self.device.set_enable(name, state)
			self.config.enable[name] = state
			self.config.save()

	def on_rate(self, obj):
		name  = Gtk.Buildable.get_name(obj)
		name  = name.replace('_spin', '')
		value = obj.get_value()
		if self.config.rate[name] != value:
			self.device.set_rate(name, value)
			self.config.rate[name] = value
			self.config.save()

	def on_config(self, obj):
		name = Gtk.Buildable.get_name(obj)
		if isinstance(obj, Gtk.SpinButton):
			value = obj.get_value()
		else:
			value = obj.get_text()
		if getattr(self.config, name) != value:
			setattr(self.config, name, value)
			self.config.save()

	def on_expose(self, obj, _):
		cairo   = obj.window.cairo_create()
		count   = len(self.history)
		extents = cairo.clip_extents()
		width   = extents[2] - extents[0] 
		height  = extents[3] - extents[1] 

		cairo.set_line_width(4)
		for i in range(0,len(self.history)):
			x = i*width/(count-1)
			y = ((1-self.history[i]) * 0.8 + 0.1) * height
			if i==0:
				cairo.move_to(x, y)
			else:
				cairo.line_to(x, y)
		cairo.stroke()

	def on_key(self, win, ev):
		if ev.string == 'q':
			Gtk.main_quit();

	def on_destroy(self, win):
		Gtk.main_quit()

        # Methods
	def load_config(self):
		def set_val(name, val):
			self.builder.get_object(name).set_value(val)
		def set_txt(name, val):
			self.builder.get_object(name).set_text(val)
		def set_act(name, val):
			self.builder.get_object(name).set_active(val)

		# Serial settings
		set_txt('device',   self.config.device)
		set_val('baudrate', self.config.baudrate)
		set_txt('parity',   self.config.parity)
		set_val('databits', self.config.databits)
		set_val('stopbits', self.config.stopbits)

		# Xively settings
		set_txt('feedid',   self.config.feedid)
		set_txt('apikey',   self.config.apikey)
		set_val('maxrate',  self.config.maxrate)

		# Sensor settings
		for key in list(self.config.enable.keys()):
			set_act(key+'_btn',  self.config.enable[key])
			set_val(key+'_spin', self.config.rate[key])


	def update(self, state):
		def setval(objs, vals, lbls):
			for i in range(0,len(objs)):
				if vals[i] == None:
					continue
				if lbls and lbls[i]:
					objs[i].set_text('%s: %f' % (lbls[i], vals[i]))
				else:
					objs[i].set_text('%f' % vals[i])

		setval(self.accs,  state.acc, ['X', 'Y', 'Z'])
		setval(self.mags,  state.mag, ['X', 'Y', 'Z'])
		setval(self.lgts,  state.lgt, [])
		setval(self.tchs,  state.tch, ['P', 'D'])
		setval(self.a2ds,  state.a2d, [])

		self.history = self.history[1:] + [state.a2d[0]]
		self.window.queue_draw()

	def debug(self, state):
		print('update: ' + str(state))
		print('\tacc - ' + str(state.acc))
		print('\tmag - ' + str(state.mag))
		print('\tlgt - ' + str(state.lgt))
		print('\ttch - ' + str(state.touch))
		print('\ta2d - ' + str(state.a2d))

	def timer(self):
		def check(state, name):
			btn = self.builder.get_object(name + '_btn')
			con = self.builder.get_object(name + '_con')
			img = Gtk.STOCK_YES if state else Gtk.STOCK_NO
			btn.set_active(state)
			con.set_from_stock(img, ICON_SIZE_BUTTON)

		check(self.device.running(), "serial")
		check(self.logger.running(), "xively")

		for item in self.device.process():
			self.update(item)
			self.logger.append(item)
			#self.debug(item)

		return True

	def run(self):
		self.load_config()
		self.window.show()
		GLib.timeout_add(1000/60, self.timer)
		Gtk.main()

	# Private methods
	def status(self, msg):
		if not msg:
			return
		status = self.builder.get_object("status")
		status.push(text=('Error: ' + msg), context_id=0)
